/************************************************************************
 * Copyright (C) 2006-2008 Philipp Marek.
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 ************************************************************************/

#ifndef __INFO_H__
#define __INFO_H__

#include "actions.h"

/** \file
 * \ref info command header file. */

/** The \ref info action. */
work_t info__work;

/** Printing the entries' info. */
action_t info__action;

#endif


