/************************************************************************
 * Copyright (C) 2006-2008 Philipp Marek.
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 ************************************************************************/

#ifndef __DIFF_H__
#define __DIFF_H__

/** \file
 * \ref diff action header file. */

#include "global.h"
#include "actions.h"

/** Diff command main function. */
work_t df__work;

#endif

