/************************************************************************
 * Copyright (C) 2006-2008 Philipp Marek.
 *
 * This program is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 ************************************************************************/

#ifndef __CHECKOUT_H__
#define __CHECKOUT_H__

#include "actions.h"
#include "global.h"

/** \file
 * Header file for the \ref checkout action. */

/** \ref checkout action. */
work_t co__work;

#endif
